#!/usr/bin/env python

# This file is part of Gummworld2.
#
# Gummworld2 is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gummworld2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.

# Compatible: Python 2.7, Python 3.2

"""07_tiled_tmw_with_renderer.py - An example of using a complex map from
The Mana World with the BasicMapRenderer in Gummworld2.

Big Daddy Demo, Newest

The demo includes collision detection and proper 2.5D placement of the avatar.
The technique relies on the renderer to draw the complete set of layers. All
game objects are inserted into the Fringe layer for rendering. The demo also
exposes more features to runtime experimentation than the previous demos.

You need The Mana World maps from https://www.themanaworld.org/. Copy the TMX
map file and the data directory into the gummworld2/data directory.
   Gfx dir: graphics/ -> data/
   Maps:    *.tmx     -> data/map/

Or get the smaller gummworld2_tmw.zip file and unpack it into the gummworld2/
directory.
"""

__version__ = '$Id: 07_tiled_tmw_with_renderer.py 437 2014-03-26 06:04:50Z stabbingfinger@gmail.com $'
__author__ = 'Gummbum, (c) 2011-2014'


import sys
import cProfile, pstats

import pygame
from pygame.locals import *

import paths
import gummworld2
from gummworld2 import context, data, model, geometry, toolkit, ui
from gummworld2 import Engine, State, TiledMap, BasicMapRenderer, Vec2d, Statf


class Avatar(model.Object):
    
    def __init__(self, map_pos):
        """Kind of a man-shaped guy."""
        model.Object.__init__(self)
        self.image = pygame.Surface((20, 45))
        self.rect = self.image.get_rect()
        self.hitbox = Rect(0, 0, 20, 5)
        pygame.draw.circle(self.image, Color('pink'), (10, 10), 9)
        pygame.draw.circle(self.image, Color('darkblue'), (10, 10), 10, 1)
        pygame.draw.rect(self.image, Color('lightblue'), (0, 20, 20, 25))
        pygame.draw.rect(self.image, Color('darkblue'), (0, 20, 20, 25), 1)
        self.image.set_colorkey(Color('black'))
        self.position = map_pos
        
        ## Tell renderer to use anti-interpolation on this object. If we don't
        ## do this, then the avatar will jitter when scrolling. This is usually
        ## necessary when avatar is used as the camera target.
        self.anti_interp = True
    
    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, val):
        # float pos
        self._position[:] = val
        x, y = val
        # my rect
        r = self.rect
        r.centerx = int(round(x))
        r.centery = int(round(y))
        # my hitbox
        r2 = self.hitbox
        r2.x = r.x
        r2.bottom = r.bottom


class App(Engine):
    
    def __init__(self, resolution=(1024, 768)):
        
        ## Load Tiled TMX map.
        """Looking at the map in Tiled, we see the following layers:
        tiled_map.layers[0] = Ground
        tiled_map.layers[1] = Fringe (where we'll put the avatar)
        tiled_map.layers[2] = Over
        tiled_map.layers[3] = Collision
        tiled_map.layers[4] = Object (which we won't do anything with)
        
        We just need to put our dynamic game objects in the right layer,
        Fringe. Currently we only have the avatar.
        """
        try:
            tiled_map = TiledMap(data.filepath('map', '001-1.tmx'))
        except:
            print('Sorry! Gummworld2 cannot include The Mana World assets' +
                  ' due to licensing.')
            print('To use this demo please download gummworld2_tmw.zip and' +
                  ' follow its instructions.')
            sys.exit()
        ## Save special layers for easy access.
        self.avatar_group = tiled_map.layers[1]
        self.collision_group = tiled_map.layers[3]
        ## Tell the renderer this layer needs to be sorted, and how to.
        self.avatar_group.objects.sort_key = lambda o: o.rect.bottom
        ## Hide the busy Collision layer. Player can show it by hitting K_3.
        self.collision_group.visible = False
        
        ## The avatar. Note: this is also the camera target.
        self.avatar = Avatar((650, 690))
        
        resolution = Vec2d(resolution)
        caption = ('07 The Mana World Map -' +
                   ' 0-9: Layer |' +
                   ' Shift 1-9: Renderer Tile Size |' +
                   ' w: Clock Wait |' +
                   ' Space: Max FPS += 60 |')
        
        Engine.__init__(self, caption=caption, resolution=resolution, camera_target=self.avatar, map=tiled_map,
                        frame_speed=0)
        
        # Conserve CPU.
        State.clock.use_wait = True
        
        ## Insert avatar into the Fringe layer.
        self.avatar_group.add(self.avatar)
        
        # A dummy to try moving before updating the avatar.
        self.collision_dummy = Avatar((0, 0))
        
        ## The renderer.
        self.renderer = BasicMapRenderer(
            tiled_map, max_scroll_speed=State.speed)
        ## New requirement. When renderer draws dynamic layers (e.g. Fringe)
        ## we need to tell it to redraw the changed tiles. This also is done
        ## in the draw cycle; see self.draw_renderer().
        self.dirty_rect = Rect(self.avatar.rect)
        self.renderer.set_dirty(self.dirty_rect)
        
        # I like huds.
        ui.text_color = Color('black')
        toolkit.make_hud()
        State.hud.add('Max FPS',
                      Statf(State.hud.next_pos(), 'Max FPS {:.0f}',
                            callback=lambda: (State.clock.max_fps,), interval=1.0))
        State.hud.add('Use Wait',
                      Statf(State.hud.next_pos(), 'Use Wait {}',
                            callback=lambda: (State.clock.use_wait,), interval=1.0))
        State.hud.add('Tile Size',
                      Statf(State.hud.next_pos(), 'Tile Size {} pixels (key={})',
                            callback=lambda: (self.renderer.tile_size, State.camera.rect.w // self.renderer.tile_size),
                            interval=1.0))

        # Create a speed box for converting mouse position to destination
        # and scroll speed.
        self.speed_box = geometry.Diamond(0, 0, 4, 2)
        self.speed_box.center = Vec2d(State.camera.rect.size) // 2
        self.max_speed_box = float(self.speed_box.width) / 3.33
        
        # Mouse and movement state. move_to is in world coordinates.
        self.move_to = None
        self.speed = None
        self.target_moved = (0, 0)
        self.mouse_down = False
        
        State.speed = 3.33
    
    def update(self, dt):
        """overrides Engine.update"""
        # If mouse button is held down update for continuous walking.
        if self.mouse_down:
            self.update_mouse_movement(pygame.mouse.get_pos())
        self.update_camera_position()
        State.camera.update()
        State.hud.update(dt)
    
    def update_mouse_movement(self, pos):
        # Angle of movement.
        angle = geometry.angle_of(self.speed_box.center, pos)
        # Final destination.
        self.move_to = None
        for edge in self.speed_box.edges:
            # line_intersects_line() returns False or (True,(x,y)).
            cross = geometry.line_intersects_line(edge, (self.speed_box.center, pos))
            if cross:
                x, y = cross[0]
                self.move_to = State.camera.screen_to_world(pos)
                self.speed = geometry.distance(self.speed_box.center, (x, y)) / self.max_speed_box
                break
    
    def update_camera_position(self):
        """update the camera's position if any movement keys are held down
        """
        if self.move_to is not None:
            camera = State.camera
            wx, wy = camera.position
            # Speed formula.
            speed = self.speed * State.speed
            # If we're within spitting distance, then taking a full step would
            # overshoot the desired destination. Therefore, we'll jump there.
            dist = geometry.distance((wx, wy), self.move_to)
            if dist < speed:
                wx, wy = self.move_to
                self.move_to = None
            else:
                # Otherwise, calculate the full step.
                angle = geometry.angle_of((wx, wy), self.move_to)
                wx, wy = geometry.point_on_circumference((wx, wy), speed, angle)
            # Check collision with environment.
            avatar = self.avatar
            dummy = self.collision_dummy
            dummy.position = wx, wy
            get_objects = self.collision_group.objects.intersect_objects
            ## Collision handling.
            hits = get_objects(dummy.hitbox)
            if not hits:
                pass
            else:
                # Try move x
                dummy.position = wx, avatar.position.y
                hits = get_objects(dummy.hitbox)
                if not hits:
                    wy = avatar.position.y
                else:
                    # Try move y
                    dummy.position = avatar.position.x, wy
                    hits = get_objects(dummy.hitbox)
                    if not hits:
                        wx = avatar.position.x
                    else:
                        # Can't move
                        self.move_to = None
                        return
            # Keep avatar inside map bounds.
            rect = State.world.rect
            wx = max(min(wx, rect.right), rect.left)
            wy = max(min(wy, rect.bottom), rect.top)
            camera.position = wx, wy
            self.avatar_group.add(avatar)
    
    def toggle_layer(self, i):
        """toggle visibility of layer i"""
        try:
            layer = State.map.layers[i]
            layer.visible = not layer.visible
            self.renderer.clear()
        except:
            pass
    
    def draw(self, interp):
        """overrides Engine.draw"""
        # Draw stuff.
        State.screen.clear()
        self.draw_renderer()
        if False:
            self.draw_debug()
        State.hud.draw()
        State.screen.flip()
    
    def draw_renderer(self):
        """renderer draws map layers"""
        
        renderer = self.renderer
        
        # If panning, mark the renderer's tiles dirty where avatar is.
        panning = False
        camera = State.camera
        camera_rect = camera.rect
        dirty_rect = self.dirty_rect
        camera_center = camera_rect.center
        if camera.target.rect.center != camera_center:
            dirty_rect.center = camera_center
            renderer.set_dirty(dirty_rect)
            panning = True
        
        # Set render's rect and draw the screen.
        renderer.set_rect(center=camera_center)
        renderer.draw_tiles()
        
        # Must mark dirty rect before next call to draw(), otherwise avatar
        # leaves little artifacts behind.
        if panning:
            renderer.set_dirty(dirty_rect)
    
    def draw_debug(self):
        # draw the hitbox and speed box
        camera = State.camera
        cx, cy = camera.rect.topleft
        rect = self.avatar.hitbox
        pygame.draw.rect(camera.surface, Color('red'), rect.move(-cx, -cy))
        pygame.draw.polygon(camera.surface, Color('white'), self.speed_box.corners, 1)
    
    def on_mouse_button_down(self, pos, button):
        self.mouse_down = True
    
    def on_mouse_button_up(self, pos, button):
        self.mouse_down = False
    
    def on_key_down(self, unicode, key, mod):
        if key == K_ESCAPE:
            context.pop()
        elif K_1 <= key <= K_9:
            if mod & KMOD_SHIFT:
                num_tiles = key - K_0
                self.renderer.tile_size = State.camera.rect.w // num_tiles
            else:
                self.toggle_layer(key - K_0)
        elif key == K_w:
            State.clock.use_wait = not State.clock.use_wait
        elif key == K_SPACE:
            clock = State.clock
            if clock.max_fps <= 540:
                if clock.fps > clock.max_fps - 60:
                    clock.max_fps += 60
                else:
                    clock.max_fps = 0
            else:
                clock.max_fps = 0
    
    def on_quit(self):
        context.pop()


def main():
    app = App()
    gummworld2.run(app)


if __name__ == '__main__':
    if False:
        cProfile.run('main()', 'prof.dat')
        p = pstats.Stats('prof.dat')
        p.sort_stats('time').print_stats()
    else:
        main()
